const EventEmitter = require('events');
const Discord = require('discord.js');
const { Events, ShardEvents, WSCodes, Status } = require('../util/Constants');
const Util = Discord.Util;
const Utils = require('../util/Util');
const Event = Utils.Events;
const EventsList = [
  Event.MESSAGE,
  Event.MESSAGEREACTIONADD,
  Event.MESSAGEREACTIONREMOVE,
  Event.READY
];
const UNRECOVERABLE_CLOSE_CODES = Object.keys(WSCodes).slice(1).map(Number);
const UNRESUMABLE_CLOSE_CODES = [1000, 4006, 4007];

class Manager extends EventEmitter {
  constructor(client) {
    Object.defineProperty(this, 'client', { value: client });
    this.shards = new Collection();
    this.totalShards = this.client.options.shards.length;
    this.gateway = null;
    Object.defineProperty(this, 'shardQueue', { value: new Set(), writable: true });
    this.reconnecting = false;
    this.status = Status.IDLE;
    this.destroyed = false;
  }

  get ping() {
    const sum = this.shards.reduce((a, b) => a + b.ping, 0);
    return sum / this.shards.size;
  }

  debug(message, shard) {
    this.client.emit(Events.DEBUG, `[WS => ${shard ? `Shard ${shard.id}` : 'Manager'}] ${message}`);
  }

  async connect() {
    const invalidToken = "Bot jetonu yanlış veya yazılmamış.";
    const {
      url: gatewayURL,
      shards: recommendedShards,
      session_start_limit: sessionStartLimit,
    } = await this.client.api.gateway.bot.get().catch(() => {
      return console.error(invalidToken);
    });

    const { total, remaining } = sessionStartLimit;

    this.debug(`Getirilen Ağ Geçidi Bilgileri
    URL: ${gatewayURL}
    Önerilen Shards: ${recommendedShards}`);

    this.debug(`Oturum Sınırı Bilgisi
    Toplam: ${total}
    Kalan: ${remaining}`);

    this.gateway = `${gatewayURL}/`;

    let { shards } = this.client.options;

    if (shards === 'auto') {
      this.debug(`Discord tarafından sağlanan önerilen shard sayısı: ${recommendedShards}`);
      this.totalShards = this.client.options.shardCount = recommendedShards;
      shards = this.client.options.shards = Array.from({ length: recommendedShards }, (_, i) => i);
    }

    this.totalShards = shards.length;
    this.debug(`Shard parçaları: ${shards.join(', ')}`);
    this.shardQueue = new Set(shards.map(id => new WebSocketShard(this, id)));

    return this.createShards();
  }

   async createShards() {
    if (!this.shardQueue.size) return false;
    const [shard] = this.shardQueue;
    this.shardQueue.delete(shard);
    if (!shard.eventsAttached) {
      shard.on(ShardEvents.ALL_READY, unavailableGuilds => {
        this.client.emit(Events.SHARD_READY, shard.id, unavailableGuilds);
        if (!this.shardQueue.size) this.reconnecting = false;
        this.checkShardsReady();
      });
      shard.on(ShardEvents.CLOSE, event => {
        if (event.code === 1000 ? this.destroyed : UNRECOVERABLE_CLOSE_CODES.includes(event.code)) {
          this.client.emit(Events.SHARD_DISCONNECT, event, shard.id);
          this.debug(WSCodes[event.code], shard);
          return;
        }
        if (UNRESUMABLE_CLOSE_CODES.includes(event.code)) {
          shard.sessionId = null;
        }
        this.client.emit(Events.SHARD_RECONNECTING, shard.id);
        this.shardQueue.add(shard);
        if (shard.sessionId) {
          this.debug(`Oturum kimliği mevcut, hemen yeniden bağlanmaya çalışılıyor...`, shard);
          this.reconnect();
        } else {
          shard.destroy({ reset: true, emit: false, log: false });
          this.reconnect();
        }
      });
      shard.on(ShardEvents.INVALID_SESSION, () => {
        this.client.emit(Events.SHARD_RECONNECTING, shard.id);
      });
      shard.on(ShardEvents.DESTROYED, () => {
        this.debug('Shard yok edildi ancak WebSocket bağlantısı yoktu! Yeniden bağlanılıyor...', shard);
        this.client.emit(Events.SHARD_RECONNECTING, shard.id);

        this.shardQueue.add(shard);
        this.reconnect();
      });
      shard.eventsAttached = true;
    }
    this.shards.set(shard.id, shard);
    try {
      await shard.connect();
    } catch (error) {
      if (error.code && UNRECOVERABLE_CLOSE_CODES.includes(error.code)) {
        throw new Error(WSCodes[error.code]);
      } else if (!error || error.code) {
        this.debug('Ağ geçidine bağlanılamadı, yeniden sıralanıyor...', shard);
        this.shardQueue.add(shard);
      } else {
        throw error;
      }
    }
    if (this.shardQueue.size) {
      this.debug(`Shard Kuyruğu Boyutu: ${this.shardQueue.size}; 5 saniye devam ediyor...`);
      await Util.delayFor(5000);
      return this.createShards();
    }
    return true;
  }

  destroy() {
    if (this.destroyed) return;
    this.debug(`Manager yıkıldı:\n${new Error('MANAGER_DESTROYED').stack} tarafından arandı.`);
    this.destroyed = true;
    this.shardQueue.clear();
    for (const shard of this.shards.values()) {
      shard.destroy({ 
        closeCode: 1000, 
        reset: true, 
        emit: false, 
        log: false 
      });
    }
  }
}

module.exports = Manager;